/*
 * Copyright (c) 2008 Graham Booker
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import "CompressorCommon.h"

/*!
 * Structure to hold a node in a tree for constructing the huffman tree
 */
typedef struct HuffmanTree_s {
	NSString				*word;				/*!< @brief The word for this tree element, nil of not a leaf*/
	int						count;				/*!< @brief The occurrance count of this element and its children*/
	struct HuffmanTree_s	*left;				/*!< @brief The left subtree*/
	struct HuffmanTree_s	*right;				/*!< @brief The right subtree*/
	struct HuffmanTree_s	*next;				/*!< @brief The next element in array (until construction is complete only)*/
} HuffmanTree;

/*!
 * The text compression engine.
 *
 * This class REQUIRES certain procedures to opperate correcty:
 * - The hasSpaces property should be set based on the data being compressed (it defaults to true).
 * - All strings should be run through scanWords:
 * - completeScan should be run.
 * - All strings should again be run through compressString to retrieve the compressed data for that string.
 *
 * This class leaves the preservation of certain metadata up to the user to save as (s)he sees fit.  Certain pieces of data MUST be preserved for successful decompression:
 * - The treeCount integer.
 * - The treeArray integer array (has treeCount number of values).
 * - The wordCount integer.
 * - The wordLengths integer array (has wordCount number of values).
 * - The compressedWords data.
 */
@interface TextCompressor : NSObject {
	NSMutableDictionary		*wordCounts;			/*!< @brief The occurrance count of each word*/
#ifdef SPLIT_OFF_PUNCTUATION
	NSMutableCharacterSet	*punct;					/*!< @brief The character set of all punctuation*/
#endif
	HuffmanTree				*tree;					/*!< @brief The huffman tree in tree structure*/
	NSMutableDictionary		*compressionLookup;		/*!< @brief The dictionary of words to compressed bits*/
	unsigned int			*treeArray;				/*!< @brief The count array of the huffman tree in cannonical form*/
	unsigned int			treeCount;				/*!< @brief Element count of the tree Array*/
	NSArray					*wordOrder;				/*!< @brief The raw data of the huffman tree word list*/
	BOOL					hasSpaces;				/*!< @brief YES if spaces are used in between words, NO otherwise*/
	unsigned int			wordCount;				/*!< @brief The number of words*/
	NSData					*compressedWords;		/*!< @The word array, compressed*/
	unsigned int			*wordLengths;			/*!< @brief The length of each word in compressedWords*/
}

@property (nonatomic) BOOL hasSpaces;						/*!< @brief Determines if input data has spaces.  MUST be set by user if not true.  If false, defaults to compressing each unichar individually.*/
@property (nonatomic, readonly) unsigned int treeCount;		/*!< @brief The depth of the huffman tree*/
@property (nonatomic, readonly) unsigned int *treeArray;	/*!< @brief The treeArray integer used for reconstructing the huffman tree.  Has treeCount number of integers*/
@property (nonatomic, readonly) unsigned int wordCount;		/*!< @brief The number of unique words in the input data*/
@property (nonatomic, readonly) unsigned int *wordLengths;	/*!< @brief The lengths of each unique word use for reconstruction of the word array.  Has wordCount number of integers.*/
@property (nonatomic, readonly) NSData *compressedWords;	/*!< @brief The compressed data of the words.*/
#ifdef SPLIT_OFF_PUNCTUATION
@property (nonatomic, readonly) NSCharacterSet *punct;		/*!< @brief The set of characters considered to be punctuation.*/
#endif

/*!
 * Scan a string to update internal occurrance counts for words.
 * @param wordString The string to scan
 */
- (void)scanWords:(NSString *)wordString;

/*!
 * Signal the scan has been completed and start construction of the huffman tree.
 */
- (void)completeScan;

/*!
 * Compress a string into a huffman bitstream.
 * @param string The string to compress.
 * @return The NSData of the compressed bitstream (trailing 0 bits are removed).
 */
- (NSData *)compressString:(NSString *)string;

@end
