/*
 * Copyright (c) 2008 Graham Booker
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import "TextCompressor.h"
#import "CompressionData.h"
#import "NSData_Extensions.h"

@implementation TextCompressor

@synthesize hasSpaces, treeArray, treeCount, wordCount, compressedWords, wordLengths;
#ifdef SPLIT_OFF_PUNCTUATION
@synthesize punct;
#endif

- (id) init
{
	self = [super init];
	if(!self)
		return self;
	
	wordCounts = [[NSMutableDictionary alloc] init];
#ifdef SPLIT_OFF_PUNCTUATION
	punct = [[NSCharacterSet punctuationCharacterSet] mutableCopy];
	[punct formUnionWithCharacterSet:[NSCharacterSet whitespaceCharacterSet]];
	[punct addCharactersInString:@"\n\r"];		
#endif
	tree = NULL;
	compressionLookup = [[NSMutableDictionary alloc] init];
	hasSpaces = YES;
	
	return self;
}

static void treeFree(HuffmanTree *tree)
{
	if(tree->left != NULL)
		treeFree(tree->left);
	if(tree->right != NULL)
		treeFree(tree->right);
	free(tree);
}

- (void) dealloc
{
	[wordCounts release];
#ifdef SPLIT_OFF_PUNCTUATION
	[punct release];
#endif
	treeFree(tree);
	[compressionLookup release];
	if(treeArray != NULL)
		free(treeArray);
	[wordOrder release];
	[compressedWords release];
	if(wordLengths != NULL)
		free(wordLengths);
	[super dealloc];
}

/*!
 * Split a string into an array of individual words.  If spaces are used, each word is separated by a space, otherwise each word is an individual unichar.
 * @param str The string to split
 * @return An NSArray of NSStrings, where each element is a single word
 */
- (NSArray *)seperateStringIntoWords:(NSString *)str
{
	NSArray *words;
	if(hasSpaces)
		words = [str componentsSeparatedByString:@" "];
	else
	{
		unsigned int length = [str length];
		NSString * *chars = (NSString * *)malloc(length * sizeof(NSString *));
		int i;
		for(i=0; i<length; i++)
			chars[i] = [str substringWithRange:NSMakeRange(i, 1)];
		words = [NSArray arrayWithObjects:chars count:length];
		free(chars);
	}
	return words;
}

/*!
 * Count the occurrance of words
 * @param ret The occurance count dictionary
 * @param word The word to add to the occurance count dictionary
 */
static inline void addOccurranceToDict(NSMutableDictionary *ret, NSString *word)
{
	NSNumber *num = [ret objectForKey:word];
	if(num == nil)
		num = [NSNumber numberWithInt:1];
	else
		num = [NSNumber numberWithInt:[num intValue]+1];
	[ret setObject:num forKey:word];
}


/*!
 * Sort function that enables sorting of words in order of increasing occurrance count
 */
NSInteger wordCompare(id first, id second, void *context)
{
	NSDictionary *lookup = (NSDictionary *)context;
	
	NSNumber *firstCount = [lookup objectForKey:first];
	NSNumber *secondCount = [lookup objectForKey:second];
	return [firstCount compare:secondCount];
}

- (void)scanWords:(NSString *)wordString
{
	NSArray *words = [self seperateStringIntoWords:wordString];
	for(NSString *word in words)
	{
#ifdef SPLIT_OFF_PUNCTUATION
		//Count punctuation
		int len = [word length];
		while(len && [punct characterIsMember:[word characterAtIndex:0]])
		{
			NSString *subKey = [punctBefore stringByAppendingString:[word substringToIndex:1]];
			addOccurranceToDict(wordCounts, subKey);
			word = [word substringFromIndex:1];
			len--;
		}
		while(len && [punct characterIsMember:[word characterAtIndex:len-1]])
		{
			int index = [word length]-1;
			NSString *subKey = [punctAfter stringByAppendingString:[word substringFromIndex:index]];
			addOccurranceToDict(wordCounts, subKey);
			word = [word substringToIndex:index];
			len--;
		}
#endif
		//Count words
		addOccurranceToDict(wordCounts, word);
	}	
}

/*!
 * Construct the compressed bits lookup from the huffman tree
 * @param tree The current huffman tree
 * @param dict The bitstream lookup indexed by word
 * @param length The current number of bits for compression
 * @param bits The current bits for compression
 */
static void addHuffmanTreeToDict(HuffmanTree *tree, NSMutableDictionary *dict, unsigned int length, unsigned int bits)
{
	//Add a 0 bit to end of bits
	unsigned int newBits = bits << 1;
	if(tree->left != NULL)
		//Construct the left side
		addHuffmanTreeToDict(tree->left, dict, length + 1, newBits);
	//Change the last bit to a 1
	newBits |= 1;
	if(tree->right != NULL)
		//Construct right side
		addHuffmanTreeToDict(tree->right, dict, length + 1, newBits);
	//If there's a word here, store the compression bits for it
	if(tree->word != NULL)
	{
		CompressionData *comp = [[CompressionData alloc] initWithLength:length andBits:bits];
		[dict setObject:comp forKey:tree->word];
		[comp release];
	}
}

/*!
 * Sort function that enables the sorting of words based upon the length of their compressed bits
 */
NSInteger canonCompare(id first, id second, void *context)
{
	NSDictionary *lookup = (NSDictionary *)context;
	
	CompressionData *firstData = [lookup objectForKey:first];
	CompressionData *secondData = [lookup objectForKey:second];
	int firstLen = firstData.length;
	int secondLen = secondData.length;
	if(firstLen > secondLen)
		return NSOrderedDescending;
	else if(firstLen < secondLen)
		return NSOrderedAscending;
	return NSOrderedSame;
}

/*!
 * Convert the huffman tree to canonical form
 * @param order The current ordering of words.
 */
- (void)makeIntoCanonicalFormWithWords:(NSArray *)order
{
	//Sort words in order of legth of compressed bits
	NSMutableArray *mutWords = [order mutableCopy];
	[mutWords sortUsingFunction:canonCompare context:compressionLookup];
	unsigned int counts[32];
	memset(counts, 0, sizeof(counts));
	
	int currentBit = 0, currentLength = 0;
	for(NSString *word in mutWords)
	{
		//Get compressed bits
		CompressionData *data = [compressionLookup objectForKey:word];
		//See if the length has changed
		int shift = data.length - currentLength;
		if(shift != 0)
		{
			//Update current compressed bits by adding 0s to increase to appropriate length
			currentBit = currentBit << shift;
			currentLength += shift;
		}
		//Update count
		counts[currentLength]++;
		//Save new compressed bits
		data.bits = currentBit;
		//Advance to next compressed bits symbol
		currentBit++;
	}
	currentLength++;
	wordOrder = mutWords;
	//Save the length counts
	treeCount = currentLength;
	treeArray = (unsigned int *)malloc(currentLength * sizeof(unsigned int));
	memcpy(treeArray, counts, currentLength * sizeof(unsigned int));
}

- (void)completeScan
{
	NSMutableArray *words = [[wordCounts allKeys] mutableCopy];
	//Sort words in increasing occurrance count
	[words sortUsingFunction:wordCompare context:wordCounts];
	//Make huffman tree
	HuffmanTree *root = NULL;
	HuffmanTree *last = NULL;
	//Construct the tree, in increasing order, as a single array
	for(NSString *key in words)
	{
		HuffmanTree *next = (HuffmanTree *)malloc(sizeof(HuffmanTree));
		next->left = next->right = next->next = NULL;
		next->count = [[wordCounts objectForKey:key] intValue];
		next->word = key;
		if(root == NULL)
			//Set the root to the first element
			root = next;
		else
			//Set the previous one's next, to make the array
			last->next = next;
		last = next;
	}
	while(1 == 1)
	{
		//Left and right have the lowest counts, combine them
		HuffmanTree *left = root;
		HuffmanTree *right = root->next;
		
		//Make new parent element, and make the left and right as children
		HuffmanTree *next = (HuffmanTree *)malloc(sizeof(HuffmanTree));
		//Remove the children from the array by advancing the root past them
		root = right->next;
		
		next->left = left;
		next->right = right;
		next->word = NULL;
		//Update count
		int count = left->count + right->count;
		next->count = count;
		
		if(root == NULL)
		{
			//There were two elements left before we started, now there's one; we're done
			root = next;
			break;
		}
		//Find place to insert this element in list
		last = NULL;
		HuffmanTree *current = root;
		while(current != NULL && current->count < count)
		{
			last = current;
			current = current->next;
		}
		if(last == NULL)
			//New element goes at begining, put it there
			root = next;
		else
			//New element goes after last
			last->next = next;
		//Set the next pointer
		next->next = current;
	}
	//Set the finalized root element
	tree = root;
	//Update the bits lookup
	addHuffmanTreeToDict(tree, compressionLookup, 0, 0);
	//Make into canonical form
	[self makeIntoCanonicalFormWithWords:words];
	[words release];
	
	wordCount = [wordOrder count];
	NSMutableData *allWords = [NSMutableData data];
	wordLengths = (unsigned int *)malloc(wordCount * sizeof(unsigned int));
	int i=0;
	for(NSString *word in wordOrder)
	{
		NSData *wordData = [word dataUsingEncoding:NSUTF8StringEncoding];
		wordLengths[i] = [wordData length];
		[allWords appendData:wordData];
		i++;
	}
	
	compressedWords = [[NSData gzipCompressedBlockWithBytes:[allWords bytes] length:[allWords length]] retain];
}

/*!
 * Append a compressed word into the bits
 * @param compressionLookup The lookup dictionary from word to compressed bits
 * @param ret The resulting data
 * @param bitsSoFarPtr The pointer to the bits already writen to the next byte
 * @param bitsUsedPtr The pointer to the count of the number of bits already writen to the next byte
 */
static inline void insertWord(NSDictionary *compressionLookup, NSString *key, NSMutableData *ret, unsigned char *bitsSoFarPtr, unsigned char *bitsUsedPtr)
{
	unsigned char bitsSoFar = *bitsSoFarPtr, bitsUsed = *bitsUsedPtr;
	CompressionData *compress = [compressionLookup objectForKey:key];
	if(compress == nil)
		NSLog(@"nil!!!  This is a serious problem!");
	int newBitCount = compress.length;
	int newBits = compress.bits;
	if(newBitCount + bitsUsed > 8)
	{
		//Bits complete one or more bytes, write them
		while(newBitCount + bitsUsed > 8)
		{
			//First complete the previous byte
			int bitsToAppend = 8 - bitsUsed;
			int extraBits = newBitCount - bitsToAppend;
			bitsSoFar = (bitsSoFar << bitsToAppend) | (newBits >> extraBits);
			[ret appendBytes:&bitsSoFar length:1];
			//Now we add in remaining bits
			bitsSoFar = 0;
			bitsUsed = 0;
			newBitCount = extraBits;
			newBits = newBits  & ((1 << newBitCount) - 1);
		}
		
		//Remaining is a partial byte
		bitsSoFar = newBits;
		bitsUsed = newBitCount;
	}
	else
	{
		//Bits do not complete a byte, save it
		bitsSoFar = (bitsSoFar << newBitCount) | newBits;
		bitsUsed += newBitCount;
	}
	//Update pointers and counters
	*bitsUsedPtr = bitsUsed;
	*bitsSoFarPtr = bitsSoFar;
}

- (NSData *)compressString:(NSString *)str
{
	NSMutableData *ret = [NSMutableData data];
	unsigned char bitsSoFar = 0, bitsUsed = 0;
	
#ifdef SPLIT_OFF_PUNCTUATION
	//Make it easy to identify punctuation
	NSMutableArray *afterPuncts = [[NSMutableArray alloc] init];
#endif
	
	//Get list of words
	NSArray *words = [self seperateStringIntoWords:str];
	NSUInteger stringWordCount = [words count];
	//Store word count
	NSUInteger storedWordCount = swab32(stringWordCount);
	[ret appendBytes:&storedWordCount length:sizeof(NSUInteger)];
	for(NSString *key in words)
	{
#ifdef SPLIT_OFF_PUNCTUATION
		//Check for punctuation at beginning of word
		int len = [key length];
		while(len && [punct characterIsMember:[key characterAtIndex:0]])
		{
			//Strip off punctuation from word
			NSString *subKey = [punctBefore stringByAppendingString:[key substringToIndex:1]];
			//Insert punctuation into compressed bits
			insertWord(compressionLookup, subKey, ret, &bitsSoFar, &bitsUsed);
			key = [key substringFromIndex:1];
			len--;
			stringWordCount++;
			//Continue to check for more
		}
		//Check for punctuation at end of word
		while(len && [punct characterIsMember:[key characterAtIndex:len-1]])
		{
			//Strip off punctuation from word
			int index = [key length]-1;
			NSString *subKey = [punctAfter stringByAppendingString:[key substringFromIndex:index]];
			//Save punctuation in list (in proper order)
			[afterPuncts insertObject:subKey atIndex:0];
			key = [key substringToIndex:index];
			len--;
			stringWordCount++;
			//Continue to check for more
		}
		//Insert word (even if it's zero length)
#endif
		insertWord(compressionLookup, key, ret, &bitsSoFar, &bitsUsed);
#ifdef SPLIT_OFF_PUNCTUATION
		//Insert punctuation that follows word
		for(NSString *afterPunct in afterPuncts)
		{
			insertWord(compressionLookup, afterPunct, ret, &bitsSoFar, &bitsUsed);
		}
		//Reset for next run
		[afterPuncts removeAllObjects];
#endif
	}
	//Save of last byte (if it's not a 0)
	if(bitsUsed)
	{
		bitsSoFar = bitsSoFar << (8 - bitsUsed);
		[ret appendBytes:&bitsSoFar length:1];
	}
#ifdef SPLIT_OFF_PUNCTUATION
	//Clean up punctuation identification
	[afterPuncts release];
	//Word count is likely to have changed, update count
	storedWordCount = swab32(stringWordCount);
	[ret replaceBytesInRange:NSMakeRange(0, sizeof(NSUInteger)) withBytes:&storedWordCount length:sizeof(NSUInteger)];
#endif
	return ret;
}

@end
