/*
 * Copyright (c) 2008 Graham Booker
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import <Cocoa/Cocoa.h>

/*States that the huffman tree should store indexes into wordArray rather than the word pointers themselves.  This is not very useful and thus not used normally*/
//#define UseWordIndexInTree

/*!
 * Structure used for storing the decoding huffman tree.  Tree is constructed inside a sufficiently large array.
 */
typedef struct HuffmanDecodingTree_s {
#ifdef UseWordIndexInTree
	NSInteger						wordNumber;
#else
	NSString						*word;
#endif
	//	left isn't needed since it is always the next structure in memory
	//	struct HuffmanDecodingTree_s	*left;
	struct HuffmanDecodingTree_s	*right;
} HuffmanDecodingTree;

/*!
 * Class for performing text decompression.
 *
 * This clas REQUIRES some out of band data that the user must provide from the compressor:
 * - The compressedWordData from the compressor.
 * - The wordCount from the compressor.
 * - The wordLengths from the compressor.
 * - The treeCount from the compressor.
 * - The treeArray from the compressor.
 *
 * Additionally, for each decompression, the user must provide
 * - A compressedData block as provided by the compressor (can be truncated if incorrect decompressed data past truncation point is acceptable).
 * - An upper bound on the final string length.
 *
 * Decompression is puposfully done in two stages.
 * The first stage returns an array of strings.  The strings in this array are not constructed, but rather the exact pointers of strings from wordArray.  This is useful information in that if one is conducting a seach, the user may find the pointers of candidates in wordArray, then examine the pointes returned in this array to determine if a match is found.  One can find a match without constructing the actual string!
 * The second stage is to create a final string from the array of strings.  There are 3 functions for accomplishing this:
 * - The first uses unicode and doesn't place spaces between words.  This is useful if the compressor was told to not use spaces, and the text is a non-latin language.
 * - The second uses unicode and inserts spaces.  This is useful for text which has some non-latin characters.
 * - The third uses MacRoman encoding and inserts spaces.  This is faster than using unicode and is useful for text which is all latin characters.
 */
@interface TextDecompressor : NSObject {
	HuffmanDecodingTree		*tree;			/*!< @brief The huffman tree stored in an array*/
	NSArray					*wordArray;		/*!< @brief The array of words stored in the tree*/
}

@property (nonatomic, readonly) NSArray *wordArray;		/*!< @brief The array of words in decompression*/

/*!
 * Create a new text decompressor given the compressed word data, the word count, and word lengths as given by the compressor.  This constructs wordArray in the instance.
 * @param compressedWordData The gzip compressed data for storing all the words
 * @param wordCount The number of words in compressedWordData
 * @param wordLengths The lengths of the words in compressedWordData
 */
- (id)initWithData:(NSData *)compressedWordData wordCount:(unsigned int)wordCount wordLengths:(unsigned int *)wordLengths;

/*!
 * Instructs the instance to construct the huffman tree from the tree count and tree array as given by the compressor.  This constructs tree in the instance.
 * @param treeCount The depth of the tree.
 * @param treeArray The number of elements at each depth in the tree
 */
- (void)constructTreeWithCount:(unsigned int)treeCount array:(unsigned int *)treeArray;

#ifdef UseWordIndexInTree
/*!
 * Decompress the data into an array of indexes of words in wordArray
 * @param compressedData The compressed version of the data
 * @param totalCount (returned) The number of words in the compressed data.
 * @return An array of indexes of words.  The user must free this array.
 */
- (NSUInteger *)decompressedWordIndexArray:(NSData *)compressedData count:(NSUInteger *)totalCount;
#else
/*!
 * Decompress the data into an array of NSStrings
 * @param compressedData The compressed version of the data
 * @param totalCount (returned) The number of words in the compressed data.
 * @return An array of words as NSStrings.  The user must free this array.
 */
- (NSString * *)decompressedWordArray:(NSData *)compressedData count:(NSUInteger *)totalCount;
#endif

@end

/*!
 * Convert an array of NSStrings into a single string using unicode and not inserting spaces
 * @param array The array of strings
 * @param count The number of words in array
 * @param stringLength The length of the final string in number of characters (can be a lower bound, determines amount of memory allocated).
 */
NSString *convertStringArrayToStringUnicodeNoSpaces(NSString * *array, int count, int stringLength);

/*!
 * Convert an array of NSStrings into a single string using unicode and inserting spaces between each word
 * @param array The array of strings
 * @param count The number of words in array
 * @param stringLength The length of the final string in number of characters (can be a lower bound, determines amount of memory allocated).
 */
NSString *convertStringArrayToStringUnicode(NSString * *array, int count, int stringLength);

/*!
 * Convert an array of NSStrings into a single string using ISO Latin 1 and inserting spaces between each word.  Note: This is faster than unicode, but will destroy data if any characters in the string are not able to be fully represented in ISO Latin 1.
 * @param array The array of strings
 * @param count The number of words in array
 * @param stringLength The length of the final string in number of characters (can be a lower bound, determines amount of memory allocated).
 */
NSString *convertStringArrayToStringMacRoman(NSString * *array, int count, int stringLength);