/*
 * Copyright (c) 2008 Graham Booker
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#import "TextDecompressor.h"
#import "NSData_Extensions.h";
#import "CompressorCommon.h"

@implementation TextDecompressor

@synthesize wordArray;

- (id)initWithData:(NSData *)compressedWordData wordCount:(unsigned int)wordCount wordLengths:(unsigned int *)wordLengths
{
	self = [super init];
	if(!self)
		return self;
	
	//Decompress the data.
	NSData *unCompressedData = [compressedWordData gunzipCompressedBlock];
	//Create the word array
	NSMutableArray *newWords = [[NSMutableArray alloc] initWithCapacity:wordCount];
	int currentOffset = 0;
	for(int i=0; i<wordCount; i++)
	{
		int length = wordLengths[i];
		//Extract the characters for this word, create the string, and insert in array.
		NSData *wordData = [unCompressedData subdataWithRange:NSMakeRange(currentOffset, length)];
		NSString *word = [[NSString alloc] initWithData:wordData encoding:NSUTF8StringEncoding];
		[newWords addObject:word];
		[word release];
		currentOffset += length;
	}
	wordArray = [newWords copy];
	[newWords release];
	
	return self;
}

- (void)dealloc
{
	[wordArray release];
	free(tree);
	[super dealloc];
}


/*
 * Construct a huffman tree from canonical form using words array and counts
 */
static void constructTree(HuffmanDecodingTree *tree, NSArray *words, unsigned int *counts, int countLength)
{
	int wordIndex;
	int treeIndex = 0;
	int currentDepth = 0;
	int stack[countLength];
	
	int depth = 0;	//depth in tree / bit length
	//Get remainder of words at this bit length
	int remainingCountInDepth = counts[0];
	int wordCount = [words count];
	
	for(wordIndex = 0; ;)
	{
		//Check if we are done with this depth
		while(remainingCountInDepth == 0)
		{
			//Go to next depth until we have words again
			depth++;
			remainingCountInDepth = counts[depth];
		}
		//Increase our depth to get to the new depth
		for(; currentDepth < depth; currentDepth++)
		{
			//Push onto the stack, so we can come back to this element to go down it's right pointer
			stack[currentDepth] = treeIndex;
			//There's no word here
#ifdef UseWordIndexInTree
			tree[treeIndex].wordNumber = -1;
#else
			tree[treeIndex].word = nil;
#endif
			//Go down left pointer
			treeIndex++;
		}
		//Set the word here
#ifdef UseWordIndexInTree
		tree[treeIndex].wordNumber = wordIndex;
#else
		tree[treeIndex].word = (NSString *)CFArrayGetValueAtIndex((CFArrayRef)words, wordIndex);
#endif
		//Go to next word
		wordIndex++;
		//Check to see if we are done
		if(wordIndex == wordCount)
			break;
		//One fewer word at this depth
		remainingCountInDepth--;
		int popIndex;
		//We must pop, because we are currently at a leaf
		do
		{
			currentDepth--;
			popIndex = stack[currentDepth];
		}while(popIndex == 0xffff);  //an index of 0xffff means that the right pointer is already used, pop till it's not anymore
		//Set the right pointer as used, because we are about to use it
		stack[currentDepth] = 0xffff;
		//Advance to next unused element
		treeIndex++;
		currentDepth++;
		//Set right pointer
		tree[popIndex].right = tree + treeIndex;
	}
}

- (void)constructTreeWithCount:(unsigned int)treeCount array:(unsigned int *)treeArray
{
	//Construct the tree array with the entire tree stored within the array
	int treeElementCount = ([wordArray count] << 1) - 1;
	tree = (HuffmanDecodingTree *)malloc(sizeof(HuffmanDecodingTree) * treeElementCount);
	constructTree(tree, wordArray, treeArray, treeCount);	
}

//Decompress compressed bits into an array of word indexes
#ifdef UseWordIndexInTree
- (NSUInteger *)decompressedWordIndexArray:(NSData *)compressedData count:(NSUInteger *)totalCount
#else
- (NSString * *)decompressedWordArray:(NSData *)compressedData count:(NSUInteger *)totalCount
#endif
{
	//Get the word count
	NSUInteger count = getCountAtOffset(compressedData, 0);
	NSUInteger wordCount = 0;
	*totalCount = count;
	
	int	bitsLeft = 7;	//Number of bits unused in bits
	int index = 1;		//Index of bits
	int arrayLength = [compressedData length]-sizeof(NSUInteger);
	const unsigned char *bitArray = [compressedData bytes]+sizeof(NSUInteger);
	//Get bits and resulting words array
	unsigned char bits = *bitArray;
	//Allocate return value and current pointer in array
#ifdef UseWordIndexInTree
	NSUInteger *words = (NSUInteger *)malloc(count * sizeof(NSUInteger));
	NSUInteger *currentWord = words;
#else
	NSString * *words = (NSString * *)malloc(count * sizeof(NSString *));
	NSString * *currentWord = words;
#endif
	
	//Get words
	while(wordCount < count)
	{
		//Start at root of tree
		HuffmanDecodingTree *current = tree;
		//Walk tree until we hit a word
#ifdef UseWordIndexInTree
		while(current->wordNumber == -1)
#else
			while(current->word == nil)
#endif
			{
				//Get current bit
				int bit = bits >> bitsLeft & 1;
				if(bitsLeft == 0)
				{
					//We've used up all our bits, get the next one
					if(index == arrayLength)
						//We've used up all our data, use 0s from here on
						bits = 0;
					else
					{
						//Get next byte of bits
						bits = bitArray[index];
						index++;
					}
					//Reset number of bits left
					bitsLeft = 7;
				}
				else
					//Decrement number of bits used
					bitsLeft--;
				
				if(bit)
					//bit was a 1, go to right
					current = current->right;
				else
					//bit was a 0, go to left
					current++;
			}
		
		//Save current word in array and go to next one
#ifdef UseWordIndexInTree
		*currentWord = current->wordNumber;
#else
		*currentWord = current->word;
#endif
		currentWord++;
		wordCount++;
	}
	//return word array
	return words;
}

NSString *convertStringArrayToStringUnicodeNoSpaces(NSString * *array, int count, int stringLength)
{
	unsigned int bufferLen = CFStringGetMaximumSizeForEncoding(stringLength, kCFStringEncodingUTF16);
	UInt8 *buffer = (UInt8 *)malloc(bufferLen);
	UInt8 *current = buffer;
	NSString * *currentWord = array;
	int wordCount;
	for(wordCount=0; wordCount < count; wordCount++)
	{
		//Get the word
		NSString *word = *currentWord;
		//Insert the word
		CFIndex bytesAdded;
		CFStringGetBytes((CFStringRef)word, CFRangeMake(0, CFStringGetLength((CFStringRef)word)), kCFStringEncodingUTF16, '?', NO, current, bufferLen, &bytesAdded);
		current += bytesAdded;
		currentWord++;
	}
	//Make the final string
	CFStringRef ret = CFStringCreateWithBytesNoCopy(NULL, buffer, current-buffer, kCFStringEncodingUTF16, NO, kCFAllocatorDefault);
	return [(NSString *)ret autorelease];
}

NSString *convertStringArrayToStringUnicode(NSString * *array, int count, int stringLength)
{
	unsigned int bufferLen = CFStringGetMaximumSizeForEncoding(stringLength, kCFStringEncodingUTF16);
	UInt8 *buffer = (UInt8 *)malloc(bufferLen);
	UInt8 *current = buffer;
	BOOL spaceNeeded = NO;
	UniChar spaceChar = ' ';
	NSString * *currentWord = array;
	int wordCount;
	for(wordCount=0; wordCount < count; wordCount++)
	{
		//Get the word
		NSString *word = *currentWord;
		CFIndex wordLen = CFStringGetLength((CFStringRef)word);
		//Get first two chars to detect punctuation
		UniChar firstTwoChars[2];
		CFStringGetCharacters((CFStringRef)word, CFRangeMake(0, 2), firstTwoChars);
		UniChar firstChar = firstTwoChars[0];
		//A byte of 1 means the punctuation follows a space
		if(spaceNeeded && firstChar != 1)
		{
			*((UniChar *)current) = spaceChar;
			current+=2;
		}
		//A byte of 0 means the punctuation precedes the space
		spaceNeeded = firstChar != 0 || wordLen == 0;
		if(firstChar < 2 && wordLen > 0)
		{
			//Insert the punctuation, which is always one char
			*((UniChar *)current) = firstTwoChars[1];
			current+=2;
		}
		else
		{
			//Insert the word
			CFIndex bytesAdded;
			CFStringGetBytes((CFStringRef)word, CFRangeMake(0, wordLen), kCFStringEncodingUTF16, '?', NO, current, bufferLen, &bytesAdded);
			current += bytesAdded;
		}
		currentWord++;
	}
	//Make the final string
	CFStringRef ret = CFStringCreateWithBytesNoCopy(NULL, buffer, current-buffer, kCFStringEncodingUTF16, NO, kCFAllocatorDefault);
	return [(NSString *)ret autorelease];
}

//See above comments for how this works, just in mac roman instead of unicode
NSString *convertStringArrayToStringMacRoman(NSString * *array, int count, int stringLength)
{
	unsigned int bufferLen = CFStringGetMaximumSizeForEncoding(stringLength, kCFStringEncodingMacRoman);
	UInt8 *buffer = (UInt8 *)malloc(bufferLen);
	UInt8 *current = buffer;
	BOOL spaceNeeded = NO;
	UniChar spaceChar = ' ';
	NSString * *currentWord = array;
	int wordCount;
	for(wordCount=0; wordCount < count; wordCount++)
	{
		NSString *word = *currentWord;
		CFIndex wordLen = CFStringGetLength((CFStringRef)word);
		UniChar firstTwoChars[2];
		CFStringGetCharacters((CFStringRef)word, CFRangeMake(0, 2), firstTwoChars);
		UniChar firstChar = firstTwoChars[0];
		if(spaceNeeded && firstChar != 1)
		{
			*current = spaceChar;
			current++;
		}
		spaceNeeded = firstChar != 0 || wordLen == 0;
		if(firstChar < 2 && wordLen > 0)
		{
			*current = firstTwoChars[1];
			current++;
		}
		else
		{
			CFIndex bytesAdded;
			CFStringGetBytes((CFStringRef)word, CFRangeMake(0, wordLen), kCFStringEncodingMacRoman, '?', NO, current, bufferLen, &bytesAdded);
			current += bytesAdded;
		}
		currentWord++;
	}
	CFStringRef ret = CFStringCreateWithBytesNoCopy(NULL, buffer, current-buffer, kCFStringEncodingMacRoman, NO, kCFAllocatorDefault);
	return [(NSString *)ret autorelease];
}

@end
